%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% SERAPH Solver based on EM & GP
% By Gang Niu, Dec 2013
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function [A_Final, iterEM, iterGP] = ...
    solverManifold_EM(eta, mu, lambda, omega, X, S, D, L, A_Init, Q_Init)
% ------------------------------------------------
% Input 
%
% eta: threshold to separate S and D
% mu: regularization parameter for entropy regularization,
%     0 means never use unlabeled data,
%     -1 means automatically compute mu=#labeled/#unlabeled
% lambda: regularization parameter for trace-norm regularization
% omega: regularization parameter for manifold regularization
%
% X: dataset, N*M matrix, N samples, M features
% S: set of similar pairs, S_ij=1 if (x_i,x_j) in S
% D: set of dissimilar pairs, D_ij=1 if (x_i,x_j) in D
% L: unnormalized graph Laplacian
%
% A_Init: initial distance metric, only for the warm start of the first M-Step
%     default is M/(M+1)*I
% Q_Init: initial probability matrix of unlabeled data, Q_ij=q(1|x_i,x_j),
%     default is Q_ij=0, all unlabeled data pairs are dissimilar pairs
%
% Output
% A_Final: final distance metric
% iterEM: number of iterations of EM
% iterGP: total number of iterations of GP
% ------------------------------------------------

epsIn = 0.1;    % error-bound of GP measured by the Frobenius norm
epsOut = 0.1;   % error-bound of EM measured by the Frobenius norm
maxEM = 10;     % maximum iterations of EM
maxIter = 10;   % maximum iterations of GP per M-Step
maxTrial = 20;  % maximum trials of gradient ascent

[N M] = size(X);
if(mu==-1)
    nLabeled = sum(sum(S+D))/2;
    nUnlabeled = N*(N-1)/2-nLabeled;
    mu = nLabeled/nUnlabeled;
elseif(mu==0)
    epsIn = 0.001;
end
if(omega==-1)
    omega = 1/(N*(N-1));
end
aux = lambda*eye(M,M)+omega*X'*L*X;

if(nargin<9)
  A_Init = M/(M+1)*eye(M,M);
end
if(nargin<10)
    Q_Init = zeros(N,N);
end

iterEM = 0;
iterGP = 0;

Q = Q_Init;
A = (A_Init+A_Init')/2;

normOut = 1e10;

while((normOut>epsOut)&&(iterEM<maxEM))
    iterEM = iterEM+1;
    
    % M-Step
    B_pre = A;
    [objL objF] = computeObj_EM(eta, mu, X, S, D, aux, B_pre, Q);
    objF_pre = objF;
    
    normIn = 1e10;
    iterCount = 0;

    while((normIn>epsIn)&&(iterCount<maxIter))
        iterCount = iterCount+1;
        iterGP = iterGP+1;
        
        % compute gradient
        grad = computeGrad_EM(eta, mu, X, S, D, aux, B_pre, Q);
                      
        % choose step-size
        deltak = 0.1*M/(norm(grad,'fro')*sqrt(iterCount));
        
        trial = 0;
        while(trial<maxTrial)
            trial = trial+1;
            
            % gradient ascent
            B_new = B_pre+deltak*grad;
        
            % projection
            B_new = (B_new+B_new')/2;
            [V b] = eig(B_new);
            b = diag(b);
            b = max(b,0);
            B_new = V*diag(b)*V';
        
            [objL objF] = computeObj_EM(eta, mu, X, S, D, aux, B_new, Q);
            
            if(objF>=objF_pre)       % success
                objF_pre = objF;
%                 fprintf('objL: %2.4f, objF: %2.4f\n',objL,objF);
                break;
            else                    % fail
                deltak = deltak/2;
            end
        end
        
        if(trial==maxTrial)               % fail
            fprintf('GP update fails... iterGP = iterGP-1, iterCount = iterCount-1.\n');
            iterGP = iterGP-1;
            iterCount = iterCount-1;
            break;
        end

        normIn = norm(B_new-B_pre,'fro');
        
        B_pre = B_new;
        
    end % end of GP iterations
    
    normOut = norm(B_pre-A,'fro');
    
    objL_pre = computeObj_EM(eta, mu, X, S, D, aux, A, Q);
    objL_new = computeObj_EM(eta, mu, X, S, D, aux, B_pre, Q);
    if(objL_pre>=objL_new)           % fail
        fprintf('L decreases! iterEM = iterEM-1, iterGP = iterGP-iterCount.\n');
        iterEM = iterEM-1;
        iterGP = iterGP-iterCount;
        break;
    end
    A = B_pre;
      
    % E-Step
    Q = computeQ(eta, mu, X, S, D, A, iterEM);
    
end % end of EM iterations

A_Final = A;

end

function [objL, objF] = computeObj_EM(eta, mu, X, S, D, aux, A, Q)
% compute the values of objective functions

epsp = 1e-12;
N = size(X,1);
XAX = X*A*X';
X2 = diag(XAX);
DD = bsxfun(@plus,X2,X2')-2*XAX;
PP = 1./(1+exp(DD-eta))+epsp;
ONE = ones(N,N);
LL = zeros(N,N);
LL(S==1) = log(PP(S==1));
LL(D==1) = log(ONE(D==1)-PP(D==1));
FF = LL;
U = ONE-S-D-eye(N,N);
if(mu>0)
    LL(U==1) = mu*(PP(U==1).*log(PP(U==1))+(ONE(U==1)-PP(U==1)).*log(ONE(U==1)-PP(U==1)));
    FF(U==1) = mu*(Q(U==1).*log(PP(U==1))+(ONE(U==1)-Q(U==1)).*log(ONE(U==1)-PP(U==1)));
end
objL = sum(sum(LL))/2-trace(aux*A);
objF = sum(sum(FF))/2-trace(aux*A);

end

function grad = computeGrad_EM(eta, mu, X, S, D, aux, A, Q)
% compute the gradient matrix of F

epsp = 1e-12;
[N] = size(X,1);
XAX = X*A*X';
X2 = diag(XAX);
DD = bsxfun(@plus,X2,X2')-2*XAX;
PP = 1./(1+exp(DD-eta))+epsp;
ONE = ones(N,N);
CC = zeros(N,N);
CC(S==1) = PP(S==1)-ONE(S==1);
CC(D==1) = PP(D==1);
U = ONE-S-D-eye(N,N);
if(mu>0)
    CC(U==1) = mu*(PP(U==1)-Q(U==1));
end
grad = X'*bsxfun(@times,sum(CC,2),X)-X'*CC*X;
grad = grad-aux;

end

function Q = computeQ(eta, mu, X, S, D, A, t)
% compute the posterior distribution of unlabeled data
% type: 1 for q^soft, 2 for q^hard, 3 for q^prj

epsp = 1e-12;
N = size(X,1);
Q = zeros(N,N);
if(mu<=0)
    return;
end
XAX = X*A*X';
X2 = diag(XAX);
DD = bsxfun(@plus,X2,X2')-2*XAX;
PP = 1./(1+exp(DD-eta))+epsp;
ONE = ones(N,N);
U = ONE-S-D-eye(N,N);
flag = (U==1);
if(isinf(t))
    Q(flag) = PP(flag);
elseif(t==0)
    tmp = double(PP>0.5)+double(PP==0.5)/2;
    Q(flag) = tmp(flag);
else
    QP(flag) = PP(flag).^(1+mu/t);
    QN(flag) = (ONE(flag)-PP(flag)).^(1+mu/t);
    Q(flag) = QP(flag)./(QP(flag)+QN(flag));
end

end
