%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% SERAPH Solver based on only GP
% By Gang Niu, Oct 2013
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function [A_Final, iterGP] = ...
    solver_GP(eta, mu, lambda, X, S, D, A_Init)
% ------------------------------------------------
% Input 
%
% eta: threshold to separate S and D
% mu: regularization parameter for entropy regularization,
%     0 means never use unlabeled data,
%     -1 means automatically compute mu=#labeled/#unlabeled
% lambda: regularization parameter for trace-norm regularization
%
% X: dataset, N*M matrix, N samples, M features
% S: set of similar pairs, S_ij=1 if (x_i,x_j) in S
% D: set of dissimilar pairs, D_ij=1 if (x_i,x_j) in D
%
% A_Init: initial distance metric
%     default is M/(M+1)*I
%
% Output
% A_Final: final distance metric
% iterGP: total number of iterations of GP
% ------------------------------------------------
epsOut = 0.001;    % error-bound of GP measured by the Frobenius norm
maxIter = 100;   % maximum iterations of GP
maxTrial = 5;  % maximum trials of gradient ascent

[N M] = size(X);
if(mu==-1)
    nLabeled = sum(sum(S+D))/2;
    nUnlabeled = N*(N-1)/2-nLabeled;
    mu = nLabeled/nUnlabeled;
end

if(nargin<7)
  A_Init = M/(M+1)*eye(M,M);
end

iterGP = 0;
A_pre = (A_Init+A_Init')/2;
objL_pre = computeObj_GP(eta, mu, lambda, X, S, D, A_pre);
A_best = A_pre;
objL_best = objL_pre;
normOut = 1e10;

while((normOut>epsOut)&&(iterGP<maxIter))
    iterGP = iterGP+1;
    
	% compute gradient
	grad = computeGrad_GP(eta, mu, lambda, X, S, D, A_pre);
    
    % choose step-size
	deltak = 0.1*M/(norm(grad,'fro')*sqrt(iterGP));
    deltakk = deltak;
        
	trial = 0;
	while(trial<maxTrial)
        trial = trial+1;
        
        % gradient ascent
        A = A_pre+deltak*grad;
        
        % projection
        A = (A+A')/2;
        [V b] = eig(A);
        b = diag(b);
        b = max(b,0);
        A = V*diag(b)*V';
        
        objL = computeObj_GP(eta, mu, lambda, X, S, D, A);
            
        if(objL>=objL_pre)       % success
        	objL_pre = objL;
            break;
        else                    % fail
            deltak = deltak/2;
        end
    end

	if(trial==maxTrial)               % fail
        fprintf('GP update fails...\n');
        A = A_pre+2*deltakk*grad;
        A = (A+A')/2;
        [V b] = eig(A);
        b = diag(b);
        b = max(b,0);
        A = V*diag(b)*V';
        objL = computeObj_GP(eta, mu, lambda, X, S, D, A);
    elseif(objL>=objL_best)         % beat best
        A_best = A;
        objL_best = objL;
    end
    
    normOut = norm(A_pre-A,'fro');
    A_pre = A;
    objL_pre = objL;
    
end % end of EM iterations

A_Final = A_best;

end

function objL = computeObj_GP(eta, mu, lambda, X, S, D, A)
% compute the values of objective functions

epsp = 1e-12;
N = size(X,1);
XAX = X*A*X';
X2 = diag(XAX);
DD = bsxfun(@plus,X2,X2')-2*XAX;
PP = 1./(1+exp(DD-eta))+epsp;
ONE = ones(N,N);
LL = zeros(N,N);
LL(S==1) = log(PP(S==1));
LL(D==1) = log(ONE(D==1)-PP(D==1));
U = ONE-S-D-eye(N,N);
if(mu>0)
    LL(U==1) = mu*(PP(U==1).*log(PP(U==1))+(ONE(U==1)-PP(U==1)).*log(ONE(U==1)-PP(U==1)));
end
objL = sum(sum(LL))/2-lambda*trace(A);

end

function grad = computeGrad_GP(eta, mu, lambda, X, S, D, A)
% compute the gradient matrix of L

epsp = 1e-12;
[N M] = size(X);
XAX = X*A*X';
X2 = diag(XAX);
DD = bsxfun(@plus,X2,X2')-2*XAX;
PP = 1./(1+exp(DD-eta))+epsp;
ONE = ones(N,N);
CC = zeros(N,N);
CC(S==1) = PP(S==1)-ONE(S==1);
CC(D==1) = PP(D==1);
U = ONE-S-D-eye(N,N);
if(mu>0)
    CC(U==1) = mu*...
        (log(PP(U==1))-log(ONE(U==1)-PP(U==1))).*PP(U==1).*(PP(U==1)-ONE(U==1));
end
grad = X'*bsxfun(@times,sum(CC,2),X)-X'*CC*X;
grad = grad-lambda*eye(M,M);

end
